/*****************************************************************************
 *
 * FILE:	arpscan.c
 * DESCRIPTION:	arp scanner
 * DATE:	Thu, Feb 17 2005
 * UPDATED:	Tue, Mar 18 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MagickWorX.COM
 * URL:		http://www.MagickWorX.COM/
 * COPYRIGHT:	(c) 2005-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2005-2008 Kouichi ABE (WALL) <kouichi@MagickWorX.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: arpscan.c,v 1.2 2008/03/18 14:40:28 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#if	HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif	/* HAVE_SYS_IOCTL_H */
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#include <ifaddrs.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/bpf.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"

/*****************************************************************************
 *
 *	Program and Copyright
 *
 *****************************************************************************/
const char	Program[]   = "arpscan";
const char	Copyright[] =
	"Copyright (c) 2005-2008 Kouichi ABE (WALL), All rights reserved.";

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/* Numeric release version identifier: MNNFFPPS: major minor fix patch status */
#define ARPSCAN_VERSION_NUMBER	0x10002000L
#define ARPSCAN_VERSION		"arpscan 1.0.2"
#define ARPSCAN_VERSION_TEXT	ARPSCAN_VERSION " (2008/03/18)"

#define	ARPSCAN_RETRY_MAX	(2)
#define	ARPSCAN_SEND_TIMEOUT	(1000)	/* msec */
#define	ARPSCAN_RECV_TIMEOUT	(10)	/* msec */

struct ifinfo {
  char			name[8];
  uint8_t		ether[ETHER_ADDR_LEN];
  uint32_t		ipaddr;
  struct ether_header	eh;
};

#define	ERRMSG_IP	"IP"
#define	ERRMSG_SUBNET	"subnet"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	version(void);
static void	usage(void);
static void	cleanup(int);
static void	set_sigtrap(void);

static int	get_ifinfo(const char *, struct ifinfo *);
static int	arpscan(const char *, uint32_t, uint32_t);
static int	send_arp(netfd_t *, struct ifinfo *, uint32_t);
static int	recv_arp(const char *, uint8_t *);
static uint32_t	inet_addr_wrap(const char *, const char *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
static void
version(void)
{
  fprintf(stderr, "%s\n%s\n\n", ARPSCAN_VERSION_TEXT, Copyright);
  exit(64);
}

static void
usage(void)
{
  fprintf(stderr, "usage: %s <ifname> <ipaddr>\n\n", Program);
  fprintf(stderr, "\n    [ examples ]\n");
  fprintf(stderr, "       %s rl0 192.168.1.1\n", Program);
  fprintf(stderr, "       %s rl0 172.16.1.100-172.16.1.200\n", Program);
  fprintf(stderr, "       %s rl0 192.168.1.0/24\n", Program);
  fprintf(stderr, "\n");
  exit(64);
}

/*****************************************************************************/

static void
cleanup(signo)
	int	signo;
{
  exit(0);
}

static void
set_sigtrap(void)
{
  struct sigaction	act;

  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)cleanup;
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGILL,  &act, NULL);
}

/*****************************************************************************/

static int
get_ifinfo(ifname, ifinfo)
	const char *	ifname;	/* network interface name */
	struct ifinfo *	ifinfo;	/* local interface information (returned) */
{
  struct ifaddrs *	ifap;
  int			status = -1;

  if (getifaddrs(&ifap) == 0) {
    register struct ifaddrs *	p;

    memset(ifinfo->name, 0, sizeof(ifinfo->name));
    strlcpy(ifinfo->name, ifname, sizeof(ifinfo->name));

    for (p = ifap; p != NULL; p = p->ifa_next) {
      if (strcmp(ifname, p->ifa_name) == 0 && p->ifa_addr != NULL) {
	if (p->ifa_addr->sa_family == AF_LINK) {
	  struct sockaddr_dl *	sdl = (struct sockaddr_dl *)p->ifa_addr;

	  if (sdl->sdl_nlen > 0) {
	    memcpy(ifinfo->ether, LLADDR(sdl), sdl->sdl_alen);
	    status++;
	    memset(ifinfo->eh.ether_dhost, 0xff, ETHER_HDR_LEN);
	    memcpy(ifinfo->eh.ether_shost, LLADDR(sdl), sdl->sdl_alen);
	    ifinfo->eh.ether_type = htons(ETHERTYPE_ARP);
	  }
	}
	else if (p->ifa_addr->sa_family == AF_INET) {
	  struct sockaddr_in *	sin = (struct sockaddr_in *)p->ifa_addr;

	  memcpy(&ifinfo->ipaddr, &sin->sin_addr.s_addr,
		 sizeof(sin->sin_addr.s_addr));
	  status++;
	}
      }
    }
    freeifaddrs(ifap);
  }

  return status;
}

/*****************************************************************************/

static int
arpscan(netif, ip_lo, ip_hi)
	const char *	netif;	/* network interface */
	uint32_t	ip_lo;	/* low ip address */
	uint32_t	ip_hi;	/* high ip address */
{
  struct ifinfo	local;

  if (get_ifinfo(netif, &local) > 0) {
    static struct bpf_insn	my_filter[] = {
      BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),		// A <- ETHER_TYPE
      BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETHERTYPE_ARP, 0, 3),	// ARP?
      BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 20),
      BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ARPOP_REPLY, 0, 1),
      BPF_STMT(BPF_RET+BPF_K, sizeof(struct ether_arp) + ETHER_HDR_LEN),
      BPF_STMT(BPF_RET+BPF_K, 0),
    };
    static struct bpf_program	my_program = {
      sizeof(my_filter) / sizeof(struct bpf_insn),
      my_filter
    };
    netfd_t *	nfd;
#if	1
    uint16_t		flags	= NETFD_BPF_IMMEDIATE|NETFD_BPF_HDRCMPLT;
    netfd_time_t	timeout	= 0;
#else
    uint16_t		flags	= NETFD_BPF_HDRCMPLT;
    netfd_time_t	timeout	= ARPSCAN_RECV_TIMEOUT;
#endif

    nfd = netfd_open_bpf(local.name, flags, 0, false, timeout, &my_program);
    if (nfd != NULL) {
      size_t	bufsize;

      if (ioctl(netfd_fileno(nfd), BIOCGBLEN, &bufsize) == 0) {
	char *	buffer;

	buffer = (char *)malloc(bufsize);
	if (buffer != NULL) {
#if	DEBUG
	  time_t	start;
	  time_t	end;

	  time(&start);
#endif	/* DEBUG */
	  do {
	    if (send_arp(nfd, &local, ip_lo) == 0) {
	      ssize_t	n;
	      int	retry = 0;

	    again:
	      memset(buffer, 0, bufsize);
	      errno = 0;
	      n = netfd_read(nfd, buffer, bufsize, ARPSCAN_RECV_TIMEOUT);
	      if (n > 0) {
		struct bpf_hdr *	bhp;
		uint8_t *		pktbdy;

		bhp    = (struct bpf_hdr *)buffer;
		pktbdy = (unsigned char *)(buffer + bhp->bh_hdrlen);
		recv_arp(local.name, pktbdy);
	      }
	      else if (n < 0 && errno == ETIMEDOUT &&
		       retry++ < ARPSCAN_RETRY_MAX) {
		goto again;
	      }
	    }
	    ip_lo = htonl(ntohl(ip_lo) + 1);
	  } while (ntohl(ip_lo) <= ntohl(ip_hi));
	  free((char *)buffer);
#if	DEBUG
	  time(&end);
	  fprintf(stderr, "total: %ld secs.\n", end - start);
#endif	/* DEBUG */

	  return 0;
	}
      }
    }
  }

  return -1;
}

static int
send_arp(nfd, local, ipaddr)
	netfd_t *	nfd;
	struct ifinfo *	local;
	uint32_t	ipaddr;	/* target ip address */
{
  unsigned char	pkt[ETHER_HDR_LEN + sizeof(struct ether_arp)];
  struct ether_arp *	arp;

  memset(pkt, 0, sizeof(pkt));
  memcpy(pkt, &local->eh, ETHER_HDR_LEN);

  arp = (struct ether_arp *)(pkt + ETHER_HDR_LEN);
  arp->arp_hrd = htons(ARPHRD_ETHER);	/* ethernet hardware format */
  arp->arp_pro = htons(ETHERTYPE_IP);	/* format of protocol address */
  arp->arp_hln = ETHER_ADDR_LEN;	/* length of hardware address */
  arp->arp_pln = sizeof(ipaddr);	/* length of protocol address */
  arp->arp_op  = htons(ARPOP_REQUEST);	/* request to resolve address */

  memcpy(&arp->arp_sha, &local->ether, sizeof(local->ether));
  memcpy(&arp->arp_spa, &local->ipaddr, sizeof(local->ipaddr));
  memcpy(&arp->arp_tpa, &ipaddr, sizeof(ipaddr));

  if (netfd_write(nfd, pkt, sizeof(pkt), ARPSCAN_SEND_TIMEOUT) == -1) {
#if	DEBUG
    fprintf(stderr, "DEBUG[send_arp] errno=%d\n", errno);
#endif	/* DEBUG */
    return -1;
  }

  return 0;
}

static int
recv_arp(netif, pktbdy)
	const char *	netif;
	uint8_t *	pktbdy;
{
  struct ether_arp *	ether_arp;

  ether_arp = (struct ether_arp *)(pktbdy + ETHER_HDR_LEN);
  if (ether_arp->arp_op == ntohs(ARPOP_REPLY)) {
    uint8_t	ether_str[NETFD_ETHER_ADDR_SIZE];
    uint8_t	ipaddr_str[NETFD_IP_ADDR_SIZE];

    snprintf(ether_str, sizeof(ether_str), "%02x:%02x:%02x:%02x:%02x:%02x",
	     ether_arp->arp_sha[0], ether_arp->arp_sha[1],
	     ether_arp->arp_sha[2], ether_arp->arp_sha[3],
	     ether_arp->arp_sha[4], ether_arp->arp_sha[5]);
    snprintf(ipaddr_str, sizeof(ipaddr_str), "%d.%d.%d.%d",
	     ether_arp->arp_spa[0], ether_arp->arp_spa[1],
	     ether_arp->arp_spa[2], ether_arp->arp_spa[3]);
    fprintf(stdout, "[%s]\t%s\t%s\n", netif, ether_str, ipaddr_str);

    return 0;
  }

  return -1;
}

/*****************************************************************************/

static uint32_t
inet_addr_wrap(addr, mesg)
	const char *	addr;
	const char *	mesg;
{
  uint32_t	ip;

  ip = inet_addr(addr);
  if (ip == INADDR_NONE) {
    fprintf(stderr, "Invalid %s address: %s\n", mesg, addr);
    usage();
  }

  return ip;
}

int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  register int	ch;
  char *	cp;
  uint32_t	ip_lo;	/* low ip address */
  uint32_t	ip_hi;	/* high ip address */
  int		status = -1;

  while ((ch = getopt(argc, argv, "hv?")) != -1) {
    switch (ch) {
      case 'v': version();
      default:	usage();
    }
  }
  argc -= optind;
  argv += optind;
  if (argc < 2) { usage(); }

  /* process arguments */
  if ((cp = strchr(argv[1], '-')) != NULL) {
    *cp++ = '\0';
    ip_lo = inet_addr_wrap(argv[1], ERRMSG_IP);
    ip_hi = inet_addr_wrap(cp, ERRMSG_IP);
    if (ntohl(ip_lo) > ntohl(ip_hi)) {
      fprintf(stderr, "Invalid address range: %s-%s\n", argv[1], cp);
      usage();
    }
  }
  else if ((cp = strchr(argv[1], '/')) != NULL) {
    uint32_t	subnet;
    uint32_t	netmask;
    int		bitmask;

    *cp++ = '\0';
    subnet  = inet_addr_wrap(argv[1], ERRMSG_SUBNET);
    subnet  = ntohl(subnet);
    bitmask = (int)strtol(cp, (char **)NULL, 10);
    if (bitmask < 1 || bitmask > 32) {
      fprintf(stderr, "Invalid netmask: %s\n", cp);
      usage();
    }
    netmask = 0xffffffff << (32 - bitmask);
    ip_lo   = htonl(subnet & netmask);
    ip_hi   = htonl(subnet | ~netmask);
  }
  else {
    ip_lo = inet_addr_wrap(argv[1], ERRMSG_IP);
    ip_hi = ip_lo;
  }

  if (netfd_init(0) == 0) {
    set_sigtrap();
    status = arpscan(argv[0], ip_lo, ip_hi);
    netfd_finish();
  }

  return status;
}
