/*****************************************************************************
 *
 * FILE:	ping0.c
 * DESCRIPTION:	fast ping for multile hosts
 * DATE:	Wed, Jun 11 2003
 * UPDATED:	Mon, Mar 17 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: ping0.c,v 1.2 2008/03/18 13:31:39 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	PING_RETRY_MAX		(5)	/* count */
#define	PING_TIMEOUT		(60)	/* seconds */
#define	PING_LOSS_RATIO(r,s)	((1.0 - ((double)(r) / (double)(s))) * 100)

typedef struct _Target {
  const char *		host;
  uint16_t		id;
  struct addrinfo *	ai;
  /**/
  uint32_t		nsend;
  uint32_t		nrecv;
  double		loss;
  uint8_t		ttl;
  double		rtt;
  struct _Target *	next;
  struct _Target *	prev;
} Target;

#define	append_target(e,l)	\
	do {	\
		(e)->next = (l);	\
		(e)->prev = (l)->prev;	\
		(l)->prev->next = (e);	\
		(l)->prev = (e);	\
	} while (0)
#define	remove_target(e,l)	\
	do {	\
		(e)->prev->next = (e)->next;	\
		(e)->next->prev = (e)->prev;	\
		append_target((e),(l)); \
		(e)->nsend = 0; \
		(e)->nrecv = 0; \
		(e)->loss = 0; \
		(e)->ttl = 0; \
		(e)->rtt = 0; \
		hostRest--;	\
	} while (0)
#define	flush_target(e,l)	\
	do {	\
		(e)->prev->next = (e)->next;	\
		(e)->next->prev = (e)->prev;	\
		freeaddrinf((e)->ai); \
		free((e)); \
	} while (0)

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	proc_packet(uint8_t *, ssize_t, uint16_t, struct sockaddr *,
			    struct timeval *);
static Target *	search_target(uint16_t);
static void	tv_sub(struct timeval *, struct timeval *);
static void	print_info(void);

static void	pinger(void);
static int	reader(const char *);
static int	lockon(const char *, uint16_t);

static void	sig_alarm(int);
static void	sig_exit(int);
static void	set_sigtrap(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Target		workList[2];
static const Target *	hostList;	/* pointer to workList */
uint8_t			flipflop = 0;	/* switch for workList array */
static int		hostMax;
static int		hostRest;
static netfd_t *	nfd = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
static int
proc_packet(ptr, len, check_id, from, tvrecv)
	uint8_t *		ptr;
	ssize_t			len;
	uint16_t		check_id;
	struct sockaddr *	from;
	struct timeval *	tvrecv;
{
  const struct ip *	ip;
  int			hlen;
  const struct icmp *	icmp;
  int			icmplen;

  ip	= (const struct ip *)ptr;		/* start of IP header */
  hlen	= ip->ip_hl << 2;			/* length of IP header */
  icmp	= (const struct icmp *)(ptr + hlen);	/* start of ICMP header */
  icmplen = len - hlen;
  if (icmplen < 8) {
    fprintf(stderr, "icmplen (%d) < 8\n", icmplen);
    return -1;
  }

  if (icmp->icmp_type == ICMP_ECHOREPLY) {
    register Target *	tp;

    if (icmplen < 16) {
      fprintf(stderr, "icmplen (%d) < 16\n", icmplen);
      return -1;
    }

    tp = search_target(icmp->icmp_id);
    if (tp) {
      struct timeval *	tvsend;
      double		rtt;

      tvsend = (struct timeval *)icmp->icmp_data;
      tv_sub(tvrecv, tvsend);
      rtt = tvrecv->tv_sec * 1000.0 + tvrecv->tv_usec / 1000.0;
      tp->nrecv++;
      tp->ttl = ip->ip_ttl;
      tp->rtt = rtt;
#if	PING_ALIKE
      fprintf(stdout, "%d bytes from %s: seq=%u, ttl=%d, rtt=%.3f ms\n",
	      icmplen, inet_ntoa(ip->ip_src), icmp->icmp_seq, ip->ip_ttl, rtt);
#endif	/* PING_ALIKE */
      fprintf(stdout, "%s: %s (ttl=%d rtt=%.3f ms loss=%.1f%%)\n",
      		tp->host, tp->nrecv > 0 ? "alive" : "dead", tp->ttl, tp->rtt,
		PING_LOSS_RATIO(tp->nrecv, tp->nsend));
      remove_target(tp, &workList[flipflop]);
    }
  }
  else if (icmp->icmp_type == ICMP_UNREACH) {
    struct icmp *	icmp_org = (struct icmp *)(icmp->icmp_data + hlen);
    register Target *	tp;

    tp = search_target(icmp_org->icmp_id);
    if (tp) {
      switch (icmp->icmp_code) {
	case ICMP_UNREACH_NET:
	  fprintf(stdout, "%s: net unreachable\n", tp->host);
	  break;
	case ICMP_UNREACH_HOST:
	  fprintf(stdout, "%s: host unreachable\n", tp->host);
	  break;
	case ICMP_UNREACH_PROTOCOL:
	  fprintf(stdout, "%s: protocol unreachable\n", tp->host);
	  break;
	case ICMP_UNREACH_PORT:
	  fprintf(stdout, "%s: port unreachable\n", tp->host);
	  break;
	case ICMP_UNREACH_NEEDFRAG:
	  fprintf(stdout, "%s: fragmentation need and DF set\n", tp->host);
	  break;
	case ICMP_UNREACH_SRCFAIL:
	  fprintf(stdout, "%s: source route failed\n", tp->host);
	  break;
	case ICMP_UNREACH_NET_UNKNOWN:	/* rfc 1812 */
	  fprintf(stdout, "%s: destination network unknown\n", tp->host);
	  break;
	case ICMP_UNREACH_HOST_UNKNOWN:	/* rfc 1812 */
	  fprintf(stdout, "%s: destination host unknown\n", tp->host);
	  break;
	case ICMP_UNREACH_ISOLATED:	/* rfc 1122 */
	  fprintf(stdout, "%s: source host isolated\n", tp->host);
	  break;
	case ICMP_UNREACH_NET_PROHIB:	/* rfc 1122 */
	  fprintf(stdout, "%s: communication with destination network administratively prohibited\n", tp->host);
	  break;
	case ICMP_UNREACH_HOST_PROHIB:	/* rfc 1122 */
	  fprintf(stdout, "%s: communication with destination host administratively prohibited\n", tp->host);
	  break;
	case ICMP_UNREACH_TOSNET:	/* rfc 1812 */
	  fprintf(stdout, "%s: network unreachable for type of service\n", tp->host);
	  break;
	case ICMP_UNREACH_TOSHOST:	/* rfc 1812 */
	  fprintf(stdout, "%s: host unreachable for type of service\n", tp->host);
	  break;
	case ICMP_UNREACH_FILTER_PROHIB:	/* rfc 1812 */
	  fprintf(stdout, "%s: communication administratively prohibited\n", tp->host);
	  break;
	case ICMP_UNREACH_HOST_PRECEDENCE:	/* rfc 1812 */
	  fprintf(stdout, "%s: host precedence violation\n", tp->host);
	  break;
	case ICMP_UNREACH_PRECEDENCE_CUTOFF:	/* rfc 1812 */
	  fprintf(stdout, "%s: precedence cutoff in effect\n", tp->host);
	default:
	  break;
      }
      remove_target(tp, &workList[flipflop]);
    }
  }
  else if (icmp->icmp_type == ICMP_SOURCEQUENCH) {
    struct icmp *	icmp_org = (struct icmp *)(icmp->icmp_data + hlen);
    register Target *	tp;

    tp = search_target(icmp_org->icmp_id);
    if (tp) {
      fprintf(stdout, "%s: packet lost, slow down\n", tp->host);
      remove_target(tp, &workList[flipflop]);
    }
  }
  else if (icmp->icmp_type == ICMP_REDIRECT) {
    struct icmp *	icmp_org = (struct icmp *)(icmp->icmp_data + hlen);
    register Target *	tp;

    tp = search_target(icmp_org->icmp_id);
    if (tp) {
      switch (icmp->icmp_code) {
	case ICMP_REDIRECT_NET:
	  fprintf(stdout, "%s: redirect datagrams for the network\n", tp->host);
	  break;
	case ICMP_REDIRECT_HOST:
	  fprintf(stdout, "%s: redirect datagrams for the host\n", tp->host);
	  break;
	case ICMP_REDIRECT_TOSNET:
	  fprintf(stdout, "%s: redirect datagrams for the type of service and network\n", tp->host);
	  break;
	case ICMP_REDIRECT_TOSHOST:
	  fprintf(stdout, "%s: redirect datagrams for the type of service and host\n", tp->host);
	default:
	  break;
      }
      remove_target(tp, &workList[flipflop]);
    }
  }
  else if (icmp->icmp_type == ICMP_TIMXCEED) {
    struct icmp *	icmp_org = (struct icmp *)(icmp->icmp_data + hlen);
    register Target *	tp;

    tp = search_target(icmp_org->icmp_id);
    if (tp) {
      switch (icmp->icmp_code) {
	case ICMP_TIMXCEED_INTRANS:
	  fprintf(stdout, "%s: time to live exceed in transit\n", tp->host);
	  break;
	case ICMP_TIMXCEED_REASS:
	  fprintf(stdout, "%s: fragment reassembly time exceeded\n", tp->host);
	default:
	  break;
      }
      remove_target(tp, &workList[flipflop]);
    }
  }
  else {
    fprintf(stderr, "unsuported: icmp type: %d code: %d\n",
	    icmp->icmp_type, icmp->icmp_code);
  }

  return 0;
}

static Target *
search_target(id)
	uint16_t	id;
{
  register Target *	tp;

  for (tp = hostList->next; tp != hostList; tp = tp->next) {
    if (id == tp->id) {
      return tp;
    }
  }

  return NULL;
}

static void
tv_sub(out, in)
	register struct timeval *	out;
	register struct timeval *	in;
{
  out->tv_usec -= in->tv_usec;
  if (out->tv_usec < 0) {
    --out->tv_sec;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}

static void
print_info(void)
{
  register Target *	tp;

  for (tp = hostList->next; tp != hostList; tp = tp->next) {
//    if (tp->nrecv == 0) { continue; }
    fprintf(stdout, "%s: %s (ttl=%d rtt=%.3fms loss=%.1f%%)\n",
  	    tp->host, tp->nrecv > 0 ? "alive" : "dead", tp->ttl, tp->rtt,
	    PING_LOSS_RATIO(tp->nrecv, tp->nsend));
  }
  fflush(stdout);
}

/*****************************************************************************/

static void
pinger(void)
{
  int			timeout;
  struct timeval	start;
  struct timeval	end;
  int			count;

  timeout = (PING_TIMEOUT * 1000) / hostMax;
#if	0
  fprintf(stderr, "receiving timeout: %d(ms)\n", timeout);
#endif
  gettimeofday(&start, (struct timezone *)NULL);
  flipflop = !flipflop;	/* change workList */
  while (hostRest > 0) {
    register Target *	tp;
    register Target *	tp_next;
    register int	rv;

    /* send packet */
    for (tp = hostList->next; tp != hostList; tp = tp->next) {
      ssize_t	n;

      tp_next = tp->next;
      n = netfd_icmp_send(nfd, tp->ai->ai_addr, tp->ai->ai_addrlen, tp->id,
			  tp->nsend++, timeout);
    }
    /* receive packet */
    count = PING_RETRY_MAX;
    do {
      struct sockaddr	from;
      socklen_t		fromlen = sizeof(from);

      errno = 0;
      rv = netfd_icmp_recv(nfd, &from, &fromlen, 0, timeout, proc_packet);
    } while (hostRest > 0 && rv == 0 && --count > 0);
    if (hostRest > 0) {
      timeout = (PING_TIMEOUT * 1000) / hostRest;
#if	0
      fprintf(stderr, "receiving timeout: %d(ms)\n", timeout);
#endif
    }
#if	0
    fprintf(stderr, "DEBUG[retry] rest=%d count=%d\n", hostRest, count);
#endif
  }
  gettimeofday(&end, (struct timezone *)NULL);
  tv_sub(&end, &start);
#if	0
  fprintf(stderr, "runtime: %ld.%ld\ttotal=%d rest=%d\n",
	  end.tv_sec, end.tv_usec, hostMax - hostRest, hostRest);
#endif
}

static int
reader(fname)
	const char *	fname;
{
#define	MAX_HOSTNAME_LEN	(80)
  register FILE *	fin;

  fin = fopen(fname, "r");
  if (fin != NULL) {
    pid_t	pid = getpid();
    char	buf[MAX_HOSTNAME_LEN];
    int		n = 0;

    hostMax  = 0;
    hostRest = 0;
    workList[0].next = workList[0].prev = &workList[0];
    workList[1].next = workList[1].prev = &workList[1];
    memset(buf, 0, sizeof(buf));
    while (fgets(buf, sizeof(buf), fin) != NULL) {
      register char *	p;
      register char *	host = buf;

      if (*buf == '#' || *buf == '\t' || *buf == '\n' || *buf == ' ') {
	continue;
      }
      p = strchr(buf, '\n');
      if (p) {
        *p = '\0';
      }
      lockon(host, (uint16_t)(pid + n));
      n++;
    }
    fclose(fin);
    hostList = &workList[flipflop];
    hostRest = hostMax;

    return 0;
  }

  return -1;
}

static int
lockon(host, id)
	const char *	host;
	uint16_t	id;
{
  Target *	tp;

  tp = (Target *)malloc(sizeof(Target));
  if (tp == NULL) {
    return -1;
  }
  tp->host	= strdup(host);
  tp->id	= id & 0xffff;
  if (netfd_host_serv(host, NULL, 0, 0, 0, &tp->ai) != 0) {
    fprintf(stderr, "%s: invalid host [skiped]\n", host);
    free((Target *)tp);
    return -1;
  }
  tp->nsend	= 0;
  tp->nrecv	= 0;
  tp->loss	= 0;
  tp->ttl	= 0;
  tp->rtt	= 0;
  append_target(tp, &workList[flipflop]);
  hostMax++;

  return 0;
}

/*****************************************************************************/

static void
sig_alarm(signo)
	int	signo;
{
  print_info();
  if (nfd) { netfd_close(nfd); }
  exit(0);
}

static void
sig_exit(signo)
	int	signo;
{
  fprintf(stderr, "killed by sig#%d\n", signo);
  print_info();
  if (nfd) { netfd_close(nfd); }
  exit(1);
}

static void
set_sigtrap(void)
{
  struct sigaction	act;
  struct timeval	interval = { 0L, 0L };
  struct timeval	value	 = { PING_TIMEOUT, 0L };
  struct itimerval	itv	 = { interval, value };

  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)sig_exit;
  sigaction(SIGINT , &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  sigaction(SIGILL,  &act, NULL);
  sigaction(SIGBUS,  &act, NULL);

  act.sa_handler = (void *)sig_alarm;
  sigaction(SIGALRM, &act, NULL);
  setitimer(ITIMER_REAL, &itv, NULL);
}

/*****************************************************************************/

int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  if (argc != 2) {
    fprintf(stderr, "usage: %s <hosts-file>\n", argv[0]);
    exit(64);
  }
  if (getuid() != 0) {
    fprintf(stderr, "%s can only be used by root\n", argv[0]);
    exit(1);
  }

  /* initialize thread engine */
  if (netfd_init(0) == -1) {
    perror("netfd_init");
    exit(1);
  }

  if (reader(argv[argc - 1]) == 0) {
    nfd = netfd_icmp_open();
    if (nfd != NULL) {
      setuid(getuid());	/* don't need special permissions any more */
      set_sigtrap();
      setvbuf(stdout, (char *)NULL, _IOLBF, 0);
      pinger();
      netfd_close(nfd);
    }
  }
  netfd_finish();

  return 0;
}
