/*****************************************************************************
 *
 * FILE:	netfd_client.c
 * DESCRIPTION:	TCP/UDP Client API for Easy Network I/O Library
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Tue, Mar 18 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd_client.c,v 1.1 2008/03/18 13:31:19 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <arpa/inet.h>
#include <sys/un.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <signal.h>
#include <errno.h>
#include "netfd.h"
#include "netfd-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/* default value */
#define	DEF_CONNECT_TIMEOUT	(5000)	/* msec */

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/
static char *	my_mktemp(char * path);
static void	sigalrm(int signo);
static int	netfd_client(const char * hostname, const char * service,
			     int socktype, int protocol, int flags,
			     netfd_time_t timeout, struct sockaddr ** saddr,
			     socklen_t * addrlen);

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/
static volatile sig_atomic_t	timedout = 0;	/* flag for connect() */ 

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
netfd_t *
netfd_tcp_client(hostname, service, timeout)
	const char *	hostname;
	const char *	service;
	netfd_time_t	timeout;	/* milliseconds */
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_client(hostname, service, SOCK_STREAM, IPPROTO_TCP, 0, timeout,
		    &saddr, &addrlen);
  if (sd != -1) {
    netfd_t *	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_udp_client(hostname, service, broadcast)
	const char *	hostname;
	const char *	service;
	bool		broadcast;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  int			sd	= -1;
  struct sockaddr *	saddr	= NULL;
  socklen_t		addrlen = 0;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= broadcast ? AF_INET : AF_UNSPEC;
  hints.ai_socktype	= SOCK_DGRAM;
  hints.ai_protocol	= IPPROTO_UDP;
  hints.ai_flags	= 0;

  if (getaddrinfo(hostname, service, &hints, &res) == 0) {
    struct addrinfo *	res_save;

    res_save = res;
    do {
      /* create a socket */
      sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
      if (sd != -1) {
	if (broadcast) {
	  const int	on = 1;
	  if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) == -1) {
	    goto error;
	  }
	}
	addrlen	= res->ai_addrlen;
	saddr	= malloc(res->ai_addrlen);
	if (saddr != NULL) {
	  memcpy(saddr, res->ai_addr, res->ai_addrlen);
	  break;
	}
      }
  error:
      my_close(sd);	/* ignore this one */
      sd = -1;
    } while ((res = res->ai_next) != NULL && errno == 0);
    freeaddrinfo(res_save);
  }

  if (sd != -1) {
    netfd_t *	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_udp_connect(hostname, service)
	const char *	hostname;
	const char *	service;
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_client(hostname, service, SOCK_DGRAM, IPPROTO_UDP, 0,
		    DEF_CONNECT_TIMEOUT, &saddr, &addrlen);
  if (sd != -1) {
    netfd_t *	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_udp_open2(hostname, connport, waitport, family, broadcast, to, tolen)
	const char *		hostname;	/* remote host */
	const char *		connport;	/* remote port */
	const char *		waitport;	/* local port */
	int			family;		/* see <sys/socket.h> */
	bool			broadcast;
	struct sockaddr **	to;		/* returned */
	socklen_t *		tolen;		/* returned */
{
  struct addrinfo	hints;
  struct addrinfo *	server;
  struct addrinfo *	client;
  int			sd = -1;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= broadcast ? AF_INET : family;
  hints.ai_socktype	= SOCK_DGRAM;
  hints.ai_protocol	= IPPROTO_UDP;
  hints.ai_flags	= 0;

  /* server */
  if (getaddrinfo(hostname, connport, &hints, &server) == 0) {
    struct addrinfo *	res_save;

    res_save = server;
    do {
      /* create a socket */
      sd = socket(server->ai_family, server->ai_socktype, server->ai_protocol);
      if (sd != -1) {
	const int	on = 1;

	(void)setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if (broadcast) {
	  if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) == -1) {
	    goto error;
	  }
	}
	*tolen	= server->ai_addrlen;
	*to	= malloc(server->ai_addrlen);
	if (*to != NULL) {
	  memcpy(*to, server->ai_addr, server->ai_addrlen);
	  break;
	}
      }
  error:
      my_close(sd);	/* ignore this one */
      sd = -1;
    } while ((server = server->ai_next) != NULL && errno == 0);
    freeaddrinfo(res_save);
  }

  /* client */
  if (sd != -1) {
    hints.ai_flags = AI_PASSIVE;
    if (getaddrinfo(NULL, waitport, &hints, &client) == 0) {
      struct addrinfo *	res_save;
      int		rc = -1;

      res_save = client;
      do {
	if ((rc = bind(sd, client->ai_addr, client->ai_addrlen)) == 0) {
	  break;
	}
      } while ((client = client->ai_next) != NULL && errno == 0);
      freeaddrinfo(res_save);
      if (rc == -1) {
	my_close(sd);	/* ignore this one */
	sd = -1;
      }
    }
  }

  if (sd != -1) {
    netfd_t *	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_ust_client(path)
	const char *	path;
{
  int	sd;

  sd = socket(AF_LOCAL, SOCK_STREAM, 0);
  if (sd != -1) {
    struct sockaddr_un	caddr;

    memset((char *)&caddr, 0, sizeof(caddr));
    caddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
    strlcpy(caddr.sun_path, path, sizeof(caddr.sun_path));
#else
    strncpy(caddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
    if (connect(sd, (struct sockaddr *)&caddr, sizeof(caddr)) == 0) {
      netfd_t *	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_udg_client(path)
	const char *	path;
{
  char	tmpath[] = "/tmp/.s.udg_client.XXXXXX";

  if (my_mktemp(tmpath) != NULL) {
    int	sd;

    sd = socket(AF_LOCAL, SOCK_DGRAM, 0);
    if (sd != -1) {
      struct sockaddr_un	caddr;

      memset((char *)&caddr, 0, sizeof(caddr));
      caddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
      strlcpy(caddr.sun_path, tmpath, sizeof(caddr.sun_path));
#else
      strncpy(caddr.sun_path, tmpath, strlen(tmpath));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
      if (bind(sd, (struct sockaddr *)&caddr, sizeof(caddr)) == 0) {
	struct sockaddr_un	saddr;

	memset((char *)&saddr, 0, sizeof(saddr));
	saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
	strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
	strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
	if (connect(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == 0) {
	  netfd_t *	nfd;

	  nfd = netfd_fdopen(sd);
	  if (nfd != NULL) {
	    nfd->path = strdup(tmpath);
	    return nfd;
	  }
	}
      }
      my_close(sd);
    }
  }

  return NULL;
}

/* cf. /usr/src/lib/libc/stdio/mktemp.c */
static char *
my_mktemp(path)
	char *	path;
{
  static const unsigned char	padchar[] =
	"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
  char *	retval = path;
  char *	trv;
  uint32_t	rnd;

  for (trv = path; *trv != '\0'; trv++) { continue; }
  if (--trv < path) {
    errno = EINVAL;
    return NULL;
  }

  while (trv >= path && *trv == 'X') {
    rnd = arc4random() % (sizeof(padchar) - 1);
    *trv-- = padchar[rnd];
  }

  return retval;
}

/*****************************************************************************/

static void
sigalrm(signo)
	int	signo __unused;
{
  timedout = 1;	/* set flag */
#if	DEBUG
  fprintf(stderr, "DEBUG[sigalrm]\n");
#endif	/* DEBUG */
}

static int
netfd_client(hostname, service, socktype, protocol, flags, timeout,
	     saddr, addrlen)
	const char *		hostname;
	const char *		service;
	int			socktype;
	int			protocol;
	int			flags;
	netfd_time_t		timeout;	/* milliseconds */
	struct sockaddr **	saddr;		/* returned */
	socklen_t *		addrlen;	/* returned */
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  int			sd = -1;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= socktype;
  hints.ai_protocol	= protocol;
  hints.ai_flags	= flags;

  if (getaddrinfo(hostname, service, &hints, &res) == 0) {
    struct addrinfo *	res_save;
    struct itimerval	value;	/* new */
    struct itimerval	ovalue;	/* old */
    struct sigaction	act;	/* new */
    struct sigaction	oact;	/* old */

    sigemptyset(&act.sa_mask);
    act.sa_flags	= 0;
    act.sa_handler	= (void *)sigalrm;
    sigaction(SIGALRM, &act, &oact);
    value.it_interval.tv_sec = value.it_interval.tv_usec = 0;
    value.it_value.tv_sec  = NETFD_MSEC2SEC(timeout);
    value.it_value.tv_usec = NETFD_MSEC2USEC(timeout);
    setitimer(ITIMER_REAL, &value, &ovalue);
    timedout = 0;	/* clear flag */

    res_save = res;
    do {
      /* create a socket */
      sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
      if (sd == -1) {
	continue;	/*ignore this one */
      }
      /* initiate a connection on a socket */
      errno = 0;
      if (connect(sd, res->ai_addr, res->ai_addrlen) == 0) {
	*addrlen = res->ai_addrlen;
	*saddr	  = (struct sockaddr *)malloc(res->ai_addrlen);
	if (*saddr != NULL) {
	  memcpy(*saddr, res->ai_addr, res->ai_addrlen);
	}
	break;
      }
      if (errno == EINTR) {
	if (timedout) {
	  errno = ETIMEDOUT;
#if	DEBUG
	  fprintf(stderr, "DEBUG[netfd_client] timedout\n");
#endif	/* DEBUG */
	}
      }
      my_close(sd);	/* ignore this one */
      sd = -1;
    } while ((res = res->ai_next) != NULL && errno == 0);
    freeaddrinfo(res_save);
    sigaction(SIGALRM, &oact, NULL);
    setitimer(ITIMER_REAL, &ovalue, NULL);
  }

  return sd;
}
