/*****************************************************************************
 *
 * FILE:	netfd_rawip.c
 * DESCRIPTION:	Raw IP Manipulation API for Easy Network I/O Library
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Thu, May  1 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd_rawip.c,v 1.3 2008/05/01 13:43:43 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <arpa/inet.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <errno.h>
#include "netfd.h"
#include "netfd-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/*
 * RFC 793
 * +--------+--------+--------+--------+
 * |           Source Address          |
 * +--------+--------+--------+--------+
 * |         Destination Address       |
 * +--------+--------+--------+--------+
 * |  zero  |  PTCL  | TCP/UDP Length  |
 * +--------+--------+--------+--------+
 */
struct pseudo_hdr {	/* TCP pseudo header for checksum */
  in_addr_t	saddr;
  in_addr_t	daddr;
  uint8_t	zero;
  uint8_t	protocol;
  uint16_t	length;
};
#define	SIZE_IPHDR	(sizeof(struct ip))
#define	SIZE_TCPHDR	(sizeof(struct tcphdr))
#define	SIZE_UDPHDR	(sizeof(struct udphdr))
#define	SIZE_ICMPHDR	(sizeof(struct icmp))
#define	SIZE_BPFHDR	(sizeof(struct bpf_hdr))
#define	SIZE_PSEUDOHDR	(sizeof(struct pseudo_hdr))

#define	MY_TTL(x)	((x) > 0 ? (x) : IPDEFTTL)

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
netfd_t *
netfd_open_raw_ip(use_own_ip_hdr)
	bool	use_own_ip_hdr;
{
  int	sd;

  sd = socket(PF_INET, SOCK_RAW, IPPROTO_RAW);
  if (sd != -1) {
    static const int	size   = IP_MAXPACKET;
    static const int	hincl  = 1;
    int			status = 0;

    (void)setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
    (void)setsockopt(sd, SOL_SOCKET, SO_SNDBUF, &size, sizeof(size));
    if (use_own_ip_hdr) {
      status = setsockopt(sd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl));
    }
    if (status == 0) {
      netfd_t *	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

ssize_t
netfd_build_tcp_packet(src_addr, src_port, dst_addr, dst_port,
		       flags, seqno, ackno, wsize, ip_ttl, ip_id, packet)
	const char *	src_addr;	/* source address */
	uint16_t	src_port;	/* source port */
	const char *	dst_addr;	/* destination address */
	uint16_t	dst_port;	/* destination port */
	uint8_t		flags;		/* tcp flags */
	uint32_t	seqno;		/* tcp sequence number */
	uint32_t	ackno;		/* tcp acknowledge number */
	uint16_t	wsize;		/* tcp window size */
	uint8_t		ip_ttl;		/* time to live */
	uint8_t		ip_id;		/* identification */
	uint8_t **	packet;
{
  if (packet != NULL) {
    uint8_t *	tcppacket;
    size_t	packetsize;
    size_t	tcpoptsize = 0;

    if ((flags & TH_SYN) == TH_SYN) {
      tcpoptsize = 12;	/* for timestamp option (see RFC1323) */
    }

    packetsize = SIZE_TCPHDR + tcpoptsize;
    tcppacket  = (uint8_t *)malloc(SIZE_PSEUDOHDR + packetsize);
    if (tcppacket != NULL) {
      const in_addr_t		saddr = inet_addr(src_addr);
      const in_addr_t		daddr = inet_addr(dst_addr);
      struct pseudo_hdr *	pseudo;
      struct tcphdr *		tcp;
      ssize_t			retval;

      pseudo = (struct pseudo_hdr *)tcppacket;
      tcp    = (struct tcphdr *)(tcppacket + SIZE_PSEUDOHDR);

      memset(tcppacket, 0, SIZE_PSEUDOHDR + packetsize);
      /* build pseudo header */
      memcpy(&pseudo->saddr, &saddr, sizeof(in_addr_t));
      memcpy(&pseudo->daddr, &daddr, sizeof(in_addr_t));
      pseudo->protocol = IPPROTO_TCP;
      pseudo->length   = htons(packetsize);

      /* build header */
      tcp->th_sport = htons(src_port);
      tcp->th_dport = htons(dst_port);
      tcp->th_seq   = htonl(seqno);
      tcp->th_ack   = htonl(ackno);
      tcp->th_x2    = 0;
      tcp->th_off   = (SIZE_TCPHDR >> 2) + (tcpoptsize >> 2);
      tcp->th_flags = flags;
      tcp->th_win   = htons(wsize);
      tcp->th_urp   = 0;

      /* tcp timestamp option */
      if (tcpoptsize) {
	uint8_t *	tstamp;
	time_t		now;

	tstamp = (uint8_t *)(tcppacket + SIZE_PSEUDOHDR + SIZE_TCPHDR);
	tstamp[0] = 8;
	tstamp[1] = 10;	/* 10bytes = kind + length + T1 + T2 */
	time(&now);
	memcpy(tstamp + 2, &now, 4);
	memset(tstamp + 6, 0, 4);	/* zero clear */
	tstamp[10] = 1;	/* NOP */
	tstamp[11] = 0;	/* EOL */
      }

      /* calculate checksum */
      tcp->th_sum  = my_in_cksum((uint16_t *)tcppacket,
				 SIZE_PSEUDOHDR + packetsize);

      /* build ip packet */
      retval = netfd_build_ip_packet(IPPROTO_TCP, src_addr, dst_addr,
				     tcppacket + SIZE_PSEUDOHDR, packetsize,
				     MY_TTL(ip_ttl), 0, IP_DF, ip_id, packet);
      if (retval > 0) {
	free(tcppacket);
      }

      return retval;
    }
  }

  return -1;
}

ssize_t
netfd_build_udp_packet(src_addr, src_port, dst_addr, dst_port, ip_ttl, ip_id,
		       packet)
	const char *	src_addr;	/* source address */
	uint16_t	src_port;	/* source port */
	const char *	dst_addr;	/* destination address */
	uint16_t	dst_port;	/* destination port */
	uint8_t		ip_ttl;		/* time to live */
	uint8_t		ip_id;		/* identification */
	uint8_t **	packet;
{
  if (packet != NULL) {
    uint8_t *	udppacket;
    size_t	packetsize;

    packetsize = SIZE_UDPHDR;
    udppacket  = (uint8_t *)malloc(SIZE_PSEUDOHDR + packetsize);
    if (udppacket != NULL) {
      const in_addr_t		saddr = inet_addr(src_addr);
      const in_addr_t		daddr = inet_addr(dst_addr);
      struct pseudo_hdr *	pseudo;
      struct udphdr *		udp;
      ssize_t			retval;

      pseudo = (struct pseudo_hdr *)udppacket;
      udp    = (struct udphdr *)(udppacket + SIZE_PSEUDOHDR);

      memset(udppacket, 0, SIZE_PSEUDOHDR + packetsize);
      /* build pseudo header */
      memcpy(&pseudo->saddr, &saddr, sizeof(in_addr_t));
      memcpy(&pseudo->daddr, &daddr, sizeof(in_addr_t));
      pseudo->protocol = IPPROTO_UDP;
      pseudo->length   = htons(packetsize);

      /* build header */
      udp->uh_sport = htons(src_port);
      udp->uh_dport = htons(dst_port);
      udp->uh_ulen  = htons(packetsize);

      /* calculate checksum */
      udp->uh_sum   = my_in_cksum((uint16_t *)udppacket,
				  SIZE_PSEUDOHDR + packetsize);

      /* build ip packet */
      retval = netfd_build_ip_packet(IPPROTO_UDP, src_addr, dst_addr,
				     udppacket + SIZE_PSEUDOHDR, packetsize,
				     MY_TTL(ip_ttl), 0, IP_DF, ip_id, packet);
      if (retval > 0) {
	free(udppacket);
      }

      return retval;
    }
  }

  return -1;
}

ssize_t
netfd_build_ip_packet(protocol, src_addr, dst_addr, data, datasize,
		      ip_ttl, ip_off, ip_tos, ip_id, packet)
	uint8_t		protocol;	/* protocol number */
	const char *	src_addr;	/* source address */
	const char *	dst_addr;	/* destination address */
	uint8_t *	data;		/* paylaod */
	size_t		datasize;	/* payload size */
	uint8_t		ip_ttl;		/* time to live */
	uint16_t	ip_off;		/* fragment offset field */
	uint8_t		ip_tos;		/* type of service */
	uint8_t		ip_id;		/* identification */
	uint8_t **	packet;		/* returned */
{
  if (packet != NULL) {
    size_t	packetsize;

    packetsize = SIZE_IPHDR + datasize;
    *packet    = (uint8_t *)malloc(packetsize);
    if (*packet != NULL) {
      const in_addr_t	saddr = inet_addr(src_addr);
      const in_addr_t	daddr = inet_addr(dst_addr);
      struct ip *	ip;

      memset(*packet, 0, packetsize);
      ip = (struct ip *)(*packet);

      /* build header */
      ip->ip_v	 = IPVERSION;
      ip->ip_hl	 = SIZE_IPHDR >> 2;
      ip->ip_tos = ip_tos;
      ip->ip_len = htons(packetsize);
      ip->ip_id	 = htons(ip_id);
      ip->ip_off = htons(ip_off);
      ip->ip_ttl = ip_ttl;
      ip->ip_p	 = protocol;
      ip->ip_sum = 0;		/* zero clear to recompute */
      ip->ip_src = *(const struct in_addr *)&saddr;
      ip->ip_dst = *(const struct in_addr *)&daddr;

      /* compute checksum */
      ip->ip_sum = my_in_cksum((uint16_t *)*packet, SIZE_IPHDR);

      /* copy payload */
      memcpy(*packet + SIZE_IPHDR, data, datasize);

      return packetsize;
    }
  }

  return -1;
}
