/*****************************************************************************
 *
 * FILE:	netfd_socket.c
 * DESCRIPTION:	Socket API for Easy Network I/O Library
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Tue, Mar 18 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd_socket.c,v 1.2 2008/03/18 14:39:39 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <arpa/inet.h>
#include <sys/un.h>
#include <net/if_dl.h>
#if	HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif	/* HAVE_SYS_STAT_H */
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <net/if_types.h>
#include <sys/time.h>
#include <sys/event.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"
#include "netfd-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/
static char *	etherprint(const uint8_t ea[6], char s[NETFD_ETHER_ADDR_SIZE]);

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
netfd_t *
netfd_socket(domain, type, protocol)
	int	domain;
	int	type;
	int	protocol;
{
  int	sd;

  sd = socket(domain, type, protocol);
  if (sd != -1) {
    netfd_t *	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_accept(nfd, addr, addrlen, timeout)
	netfd_t *		nfd;
	struct sockaddr *	addr;
	socklen_t *		addrlen;
	netfd_time_t		timeout;	/* milliseconds */
{
  int		sd;
  netfd_t *	newfd;
#ifdef	SO_NOSIGPIPE
  const int	on = 1;
#endif	/* SO_NOSIGPIPE */

#if	0
  while ((sd = accept(nfd->osfd, addr, addrlen)) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return NULL; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return NULL; }
  }
#else
  pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

  if (netfd_poll(&pfds, 1, timeout) <= 0) { return NULL; }
  while ((sd = accept(nfd->osfd, addr, addrlen)) < 0) {
    if (errno == EINTR)		{ continue; }
    if (!NETFD_IO_NOT_READY)	{ return NULL; }
  }
#endif

#ifdef	SO_NOSIGPIPE
  /* Do not generate SIGPIPE */
  (void)setsockopt(sd, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on));
#endif	/* SO_NOSIGPIPE */

  newfd = netfd_fdopen(sd);
  if (newfd != NULL) {
    newfd->saddr = (struct sockaddr *)malloc(*addrlen);
    if (newfd->saddr != NULL) {
      memcpy(newfd->saddr, addr, *addrlen);
      newfd->addrlen = *addrlen;
    }
    else {
      netfd_close(newfd);
      newfd = NULL;
    }
  }
  else {
    my_close(sd);
  }

  return newfd;
}

/*****************************************************************************/

ssize_t
netfd_readv(nfd, iov, iovcnt, timeout)
	netfd_t *		nfd;
	const struct iovec *	iov;
	int			iovcnt;
	netfd_time_t		timeout;	/* millisedoncs */
{
  pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };
  ssize_t	n;

  if (netfd_poll(&pfds, 1, timeout) <= 0) { return -1; }
  while ((n = readv(nfd->osfd, iov, iovcnt)) < 0) {
    if (errno == EINTR)		{ continue; }
    if (!NETFD_IO_NOT_READY)	{ return -1; }
  }

  return n;
}

ssize_t
netfd_writev(nfd, iov, iovcnt, timeout)
	netfd_t *		nfd;
	const struct iovec *	iov;
	int			iovcnt;
	netfd_time_t		timeout;	/* millisedoncs */
{
  int		i;
  ssize_t	nbytes;
  size_t	resid;
  ssize_t	n;

  for (nbytes = 0, i = 0; i < iovcnt; i++) {
    nbytes += iov[i].iov_len;
  }

  resid = nbytes;
  while (resid > 0) {
    n = writev(nfd->osfd, iov, iovcnt);
    if (n == -1) {
      pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

      if (errno == EINTR)			{ continue; }
      if (!NETFD_IO_NOT_READY)			{ return -1; }
      if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
    }
    else {
      resid -= n;
      break;
    }
  }

  if (resid > 0) {
    for (i = 0; i < iovcnt; i++) {
      if ((size_t)n >= iov[i].iov_len) {	/* written */
	n -= iov[i].iov_len;
      }
      else {
	ssize_t	l;

	l = netfd_write(nfd, iov[i].iov_base, iov[i].iov_len, timeout);
	if (l == -1) {
	  break;
	}
	n     -= l;
	resid -= l;
      }
    }
  }

  return (ssize_t)(nbytes - resid);
}

/*****************************************************************************/

ssize_t
netfd_recvfrom(nfd, buf, len, flags, from, fromlen, timeout)
	netfd_t *		nfd;
	void *			buf;
	size_t			len;
	int			flags;
	struct sockaddr *	from;
	socklen_t *		fromlen;
	netfd_time_t		timeout;	/* milliseconds */
{
  pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };
  ssize_t	n;

  if (netfd_poll(&pfds, 1, timeout) <= 0) { return -1; }

  while ((n = recvfrom(nfd->osfd, buf, len, flags, from, fromlen)) < 0) {
    if (errno == EINTR)		{ continue; }
    if (!NETFD_IO_NOT_READY)	{ return -1; }
  }

  return n;
}

ssize_t
netfd_sendto(nfd, msg, len, flags, to, tolen, timeout)
	netfd_t *		nfd;
	const void *		msg;
	size_t			len;
	int			flags;
	const struct sockaddr *	to;
	socklen_t		tolen;
	netfd_time_t		timeout;	/* milliseconds */
{
  ssize_t	n;

  while ((n = sendto(nfd->osfd, msg, len, flags, to, tolen)) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }

  return n;
}


#if	HAVE_FREEBSD_SENDFILE
int
netfd_sendfile(nfd, nsd, init_offset, nbytes, hdtr, ret_sbytes, flags, timeout)
	netfd_t *		nfd;
	netfd_t *		nsd;
	off_t			init_offset;
	size_t			nbytes;
	struct sf_hdtr *	hdtr;
	off_t *			ret_sbytes;
	int			flags;
	netfd_time_t		timeout;	/* milliseconds */
{
  off_t	offset;
  off_t	sbytes;

  offset = init_offset;
  while (sendfile(nfd->osfd, nsd->osfd, offset, nbytes, hdtr, &sbytes,
		  flags) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

    offset += sbytes;
    nbytes -= sbytes;

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }
  if (ret_sbytes != NULL) {
    *ret_sbytes = offset + sbytes - init_offset;
  }

  return 0;
}
#else
#if	HAVE_LINUX_SENDFILE
/*
 * I have no linux environments. Is this routine ok?
 */
int
netfd_sendfile(nfd, nsd, init_offset, nbytes, hdtr, ret_sbytes, flags, timeout)
	netfd_t *	nfd;
	netfd_t *	nsd;
	off_t		init_offset;
	size_t		nbytes;
	void *		hdtr;
	off_t *		ret_sbytes;
	int		flags;
	netfd_time_t	timeout;	/* milliseconds */
{
  off_t	offset;
  off_t	sbytes;

  offset = init_offset;
  while (sendfile(nfd->osfd, nsd->osfd, &sbytes, nbytes) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

    offset += sbytes;
    nbytes -= sbytes;

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }
  if (ret_sbytes != NULL) {
    *ret_sbytes = offset + sbytes - init_offset;
  }

  return 0;
}
#else	/* No sendfile(2) */
#include <sys/mman.h>
int
netfd_sendfile(nfd, nsd, init_offset, nbytes, hdtr, ret_sbytes, flags, timeout)
	netfd_t *	nfd;
	netfd_t *	nsd;
	off_t		init_offset;
	size_t		nbytes;
	void *		hdtr __unused;
	off_t *		ret_sbytes;
	int		flags __unused;
	netfd_time_t	timeout;	/* milliseconds */
{
  void *	buf;
  size_t	len;
  int		n;
  int		status = 0;

  if (nbytes == 0) {
    /* get whole file size */
    n = netfd_seek(nfd, (off_t)0, NETFD_SEEK_END);
    if (n == -1) {
      return -1;
    }
    len = (size_t)n;
  }
  else {
    len = nbytes;
  }
  buf = mmap(NULL, len, PROT_READ, MAP_PRIVATE, nfd->osfd, init_offset);
  if (buf == MAP_FAILED) {
    return -1;
  }
  n = netfd_write(nsd, buf, len, timeout);
  if (n == -1) {
    status = -1;
  }
  if (ret_sbytes != NULL) {
    *ret_sbytes = n;
  }
  munmap(buf, len);

  return status;
}
#endif	/* HAVE_LINUX_SENDFILE */
#endif	/* HAVE_FREEBSD_SENDFILE */

/*****************************************************************************/

int
netfd_host_serv(hostname, service, family, socktype, protocol, res)
	const char *		hostname;
	const char *		service;
	int			family;
	int			socktype;
	int			protocol;
	struct addrinfo **	res;
{
  struct addrinfo	hints;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;	/* always return canonical name */
  hints.ai_family	= family;
  hints.ai_socktype	= socktype;
  hints.ai_protocol	= protocol;

  return getaddrinfo(hostname, service, &hints, res);
}

int
netfd_getnameinfo(nfd, host, hostlen, serv, servlen)
	netfd_t *	nfd;
	char *		host;
	size_t		hostlen;
	char *		serv;
	size_t		servlen;
{
  struct sockaddr	name;
  socklen_t		namelen = sizeof(name);

  if (getpeername(nfd->osfd, &name, &namelen) == -1) {
    return -1;
  }
  if (getnameinfo(&name, namelen, host, hostlen, serv, servlen,
		  NI_NUMERICHOST | NI_NUMERICSERV) == -1) {
    return -1;
  }
  return 0;
}

char *
netfd_sock_ntop(sa0, with_port, dst, size)
	const struct sockaddr *	sa0;
	bool			with_port;
	char *			dst;
	size_t			size;
{
  static char	host[132];	/* Unix domain is largest */
  char *	retval = host;

  memset(host, 0, sizeof(host));
  switch (sa0->sa_family) {
    case 255: {
	unsigned long	mask;
	unsigned int	index0 = 1 << 31;
	unsigned short	new_mask = 0;

	mask = ntohl(((const struct sockaddr_in *)sa0)->sin_addr.s_addr);
	while (mask & index0) {
	  new_mask++;
	  index0 >>= 1;
	}
	if (snprintf(host, sizeof(host), "/%hu", new_mask) == -1) {
	  return NULL;
	}
      }
      break;
    case AF_UNSPEC:
    case AF_INET: {
	const struct sockaddr_in *	sa = (const struct sockaddr_in *)sa0;

	if (inet_ntop(AF_INET, &sa->sin_addr, host, sizeof(host)) != NULL) {
	  size_t	len = strlen(host);

	  if (ntohs(sa->sin_port) != 0 && with_port == true) {
	    char	port[7];
	    int		n;

	    n = snprintf(port, sizeof(port), ".%hd", ntohs(sa->sin_port));
	    if (n != -1) {
	      strlcat(host, port, sizeof(host) - len);
	    }
	  }
	}
      }
      break;
    case AF_INET6: {
	const struct sockaddr_in6 *	sa = (const struct sockaddr_in6 *)sa0;

	if (inet_ntop(AF_INET6, &sa->sin6_addr, host, sizeof(host)) != NULL) {
	  size_t	len = strlen(host);

	  if (ntohs(sa->sin6_port) != 0 && with_port == true) {
	    char	port[7];
	    int		n;

	    n = snprintf(port, sizeof(port), ".%hd", ntohs(sa->sin6_port));
	    if (n != -1) {
	      strlcat(host, port, sizeof(host) - len);
	    }
	  }
	}
      }
      break;
    case AF_UNIX: {
	const struct sockaddr_un *	sa = (const struct sockaddr_un *)sa0;

	/*
	 * OK to have no pathname bound to the socket: happens on every
	 * connect() unless client calls bind() first.
	 */
	if (sa->sun_path[0] == 0) {
	  strlcpy(host, "(no pathname bound)", sizeof(host));
	}
	else {
	  snprintf(host, sizeof(host), "%s", sa->sun_path);
	}
      }
      break;
    case AF_LINK: {
	const struct sockaddr_dl *	sa = (const struct sockaddr_dl *)sa0;

	if (sa->sdl_type == IFT_ETHER && sa->sdl_alen > 0) {
	  char	s[NETFD_ETHER_ADDR_SIZE];

	  etherprint((uint8_t *)LLADDR(sa), s);
	  strlcpy(host, s, sizeof(host));
	}
	else if (sa->sdl_nlen > 0) {
	  /* interface name */
	  memset(host, 0, sizeof(host));
	  memcpy(host, sa->sdl_data, sa->sdl_nlen);
	}
	else {
	  snprintf(host, sizeof(host), "<Link#%d>", sa->sdl_index);
	}
      }
      break;
    default:
      break;
  }

  if (dst != NULL) {
    strlcpy(dst, host, size);
  }

  return retval;
}

static char *
etherprint(ea, s)
	const uint8_t	ea[6];
	char		s[NETFD_ETHER_ADDR_SIZE];
{
  snprintf(s, NETFD_ETHER_ADDR_SIZE,
	   "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
	   ea[0], ea[1], ea[2], ea[3], ea[4], ea[5]);
  s[NETFD_ETHER_ADDR_SIZE-1] = '\0';

  return s;
}
